 MODULE advection


!------------------------------------------------------------------------------!
! Current revisions:
! -----------------
! +prho, prho_1
! +bc_lr_cyc, bc_ns_cyc
! +output_for_t0 
! translation error of actual -> current revisions fixed
! +q* in dots_label, dots_unit. increased dots_num respectively
! typographical error in dots_unit fixed
! +clip_dvrp_*, cluster_size, color_interval, dvrpsize_interval, dvrp_overlap,
! dvrp_total_overlap, groundplate_color, local_dvrserver_running, n*_dvrp,
! interval_*_dvrp_prt, isosurface_color, particle_color, particle_dvrpsize,
! topography color in dvrp_variables,
! vertical_particle_advection is a 1d-array,
! +canyon_height, canyon_width_x, canyon_width_y, canyon_wall_left, 
! canyon_wall_south, conserve_volume_flow_mode, coupling_start_time,
! dp_external, dp_level_b, dp_level_ind_b, dp_smooth, dp_smooth_factor, dpdxy,
! run_coupled, time_since_reference_point, u_bulk, v_bulk in control_parameters,
! default value of grid_matching changed to strict
! +shf_av, qsws_av
!
! Former revisions:
! -----------------
! $Id: modules.f90 388 2009-09-23 09:40:33Z raasch $
!
! 217 2008-12-09 18:00:48Z letzel
! +topography_grid_convention
! some dvrp-variables changed to single precision, variables for dvrp-mode
! pathlines added, +target_id, abort_mode, message_string
!
! 197 2008-09-16 15:29:03Z raasch
! allow 100 spectra levels instead of 10 for consistency with 
! define_netcdf_header, +canopy_heat_flux, cthf, lai,
! +leaf_surface_concentration, scalar_exchange_coefficient, sec, sls
! +hor_index_bounds, hor_index_bounds_previous_run, id_inflow, id_recycling,
! inflow_damping_*, mean_inflow_profiles, numprocs_previous_run, nx_on_file,
! ny_on_file, offset_ocean_*, recycling_plane, recycling_width, turbulent_inflow
! -myid_char_14
!
! 138 2007-11-28 10:03:58Z letzel
! +drag_coefficient, pch_index, lad_surface, lad_vertical_gradient, 
! lad_vertical_gradient_level, plant_canopy, lad, lad_s, lad_u, lad_v, 
! lad_w, cdc, lad_vertical_gradient_level_ind, canopy_mode
! +dt_sort_particles, ngp_2dh_s_inner, time_sort_particles, flags, 
! wall_flags_1..10, wall_humidityflux(0:4), wall_qflux(0:4), 
! wall_salinityflux(0:4), wall_scalarflux(0:4)
!
! 108 2007-08-24 15:10:38Z letzel
! +comm_inter, constant_top_momentumflux, coupling_char, coupling_mode,
! coupling_mode_remote, c_u, c_v, c_w, dt_coupling, e_init, humidity_remote, 
! ngp_xy, nxlu, nysv, port_name, qswst_remote, terminate_coupled, 
! terminate_coupled_remote, time_coupling, top_momentumflux_u|v, type_xy, 
! uswst*, vswst*
!
! 97 2007-06-21 08:23:15Z raasch
! +atmos_ocean_sign, ocean, r, + salinity variables
! defaults of .._vertical_gradient_levels changed from -1.0 to -9999999.9
! hydro_press renamed hyp, use_pt_reference renamed use_reference
!
! 89 2007-05-25 12:08:31Z raasch
! +data_output_pr_user, max_pr_user, size of data_output_pr, dopr_index,
! dopr_initial_index and dopr_unit enlarged,
! var_hom and var_sum renamed pr_palm
!
! 82 2007-04-16 15:40:52Z raasch
! +return_addres, return_username
! Cpp-directive lcmuk renamed lc
!
! 75 2007-03-22 09:54:05Z raasch
! +arrays precipitation_amount, precipitation_rate, precipitation_rate_av,
! rif_wall, z0_av, +arrays u_m_l, u_m_r, etc. for radiation boundary conditions,
! +loop_optimization, netcdf_64bit_3d, zu_s_inner, zw_w_inner, id_var_zusi_*,
! id_var_zwwi_*, ts_value, u_nzb_p1_for_vfc, v_nzb_p1_for_vfc, pt_reference,
! use_pt_reference, precipitation_amount_interval, revision
! +age_m in particle_type, moisture renamed humidity,
! -data_output_ts, dots_n, uvmean_outflow, uxrp, vynp,
! arrays dots_label and dots_unit now dimensioned with dots_max,
! setting of palm version moved to main program
!
! 37 2007-03-01 08:33:54Z raasch
! +constant_top_heatflux, top_heatflux, use_top_fluxes, +arrays for top fluxes,
! +nzt_diff, default of bc_pt_t renamed "initial_gradient"
! Bugfix: p is not a pointer
!
! RCS Log replace by Id keyword, revision history cleaned up
!
! Revision 1.95  2007/02/11 13:18:30  raasch
! version 3.1b (last under RCS control)
!
! Revision 1.1  1997/07/24 11:21:26  raasch
! Initial revision
!
!
! Description:
! ------------
! Definition of variables for special advection schemes
!------------------------------------------------------------------------------!

    REAL ::  spl_gamma_x, spl_gamma_y

    REAL, DIMENSION(:), ALLOCATABLE   ::  aex, bex, dex, eex, spl_z_x, spl_z_y
    REAL, DIMENSION(:,:), ALLOCATABLE ::  spl_tri_x, spl_tri_y, spl_tri_zu, &
                                          spl_tri_zw

    SAVE

 END MODULE advection




 MODULE array_kind

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of type parameters (used for the definition of single or double
! precision variables)
!------------------------------------------------------------------------------!

    INTEGER, PARAMETER ::  dpk = SELECTED_REAL_KIND( 12 ), &
                           spk = SELECTED_REAL_KIND( 6 )

    SAVE

 END MODULE array_kind




 MODULE arrays_3d

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of all arrays defined on the computational grid
!------------------------------------------------------------------------------!

    USE array_kind

    REAL, DIMENSION(:), ALLOCATABLE ::                                         &
          ddzu, dd2zu, dzu, ddzw, dzw, hyp, inflow_damping_factor, km_damp_x,  &
          km_damp_y, lad, l_grid, pt_init, q_init, rdf, sa_init, ug, u_init,   &
          u_nzb_p1_for_vfc, vg, v_init, v_nzb_p1_for_vfc, zu, zw

    REAL, DIMENSION(:,:), ALLOCATABLE ::                                       &
          c_u, c_v, c_w, dzu_mg, dzw_mg, f1_mg, f2_mg, f3_mg,                  &
          mean_inflow_profiles, pt_slope_ref, qs, qswst_remote, ts, us, z0

    REAL, DIMENSION(:,:), ALLOCATABLE, TARGET ::                               &
          qsws_1, qsws_2, qswst_1, qswst_2, rif_1, rif_2, saswsb_1, saswst_1,  &
          shf_1, shf_2, tswst_1, tswst_2, usws_1, usws_2, uswst_1, uswst_2,    &
          vsws_1, vsws_2, vswst_1, vswst_2

    REAL, DIMENSION(:,:), POINTER ::                                           &
          qsws, qsws_m, qswst, qswst_m, rif, rif_m, saswsb, saswst, shf,       &
          shf_m, tswst, tswst_m, usws, uswst, usws_m, uswst_m, vsws, vswst,    &
          vsws_m, vswst_m

    REAL, DIMENSION(:,:,:), ALLOCATABLE ::                                     &
          canopy_heat_flux, cdc, d, diss, lad_s, lad_u, lad_v, lad_w, lai,     &
          l_wall, sec, sls, tend, u_m_l, u_m_n, u_m_r, u_m_s, v_m_l, v_m_n,    &
          v_m_r, v_m_s, w_m_l, w_m_n, w_m_r, w_m_s

    REAL, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::                             &
          ql_v, ql_vp

    REAL, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::                             &
          e_1, e_2, e_3, kh_1, kh_2, km_1, km_2, p, prho_1, pt_1, pt_2, pt_3,  &
          q_1, q_2, q_3, ql_1, ql_2, rho_1, sa_1, sa_2, sa_3, u_1, u_2, u_3,   &
          v_1, v_2, v_3, vpt_1, vpt_2, w_1, w_2, w_3

    REAL, DIMENSION(:,:,:), POINTER ::                                         &
          e, e_m, e_p, kh, kh_m, km, km_m, prho, pt, pt_m, pt_p, q, q_m, q_p,  &
          ql, ql_c, rho, sa, sa_p, te_m, tpt_m, tq_m, tsa_m, tu_m, tv_m, tw_m, &
          u, u_m, u_p, v, v_m, v_p, vpt, vpt_m, w, w_m, w_p

    REAL, DIMENSION(:,:,:,:), ALLOCATABLE ::  rif_wall

    SAVE

 END MODULE arrays_3d




 MODULE averaging

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of variables needed for time-averaging of 2d/3d data
!------------------------------------------------------------------------------!

    REAL, DIMENSION(:,:), ALLOCATABLE ::  lwp_av, precipitation_rate_av,       &
                                          qsws_av, shf_av,ts_av, us_av, z0_av

    REAL, DIMENSION(:,:,:), ALLOCATABLE, TARGET :: &
          e_av, p_av, pc_av, pr_av, pt_av, q_av, ql_av, ql_c_av, ql_v_av, &
          ql_vp_av, qv_av, rho_av, s_av, sa_av, u_av, v_av, vpt_av, w_av

 END MODULE averaging




 MODULE cloud_parameters

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of variables and constants for cloud physics
!------------------------------------------------------------------------------!

    REAL  ::  b_cond, cp = 1005.0, diff_coeff_l = 0.23E-4,                     &
              effective_coll_efficiency, l_d_cp, l_d_r, l_d_rv, l_v = 2.5E+06, &
              mass_of_solute, molecular_weight_of_solute,                      &
              prec_time_const = 0.001, ql_crit = 0.0005, rho_l = 1.0E3,        &
              r_d = 287.0, r_v = 461.51, thermal_conductivity_l = 2.43E-2

    REAL, DIMENSION(:), ALLOCATABLE   ::  pt_d_t, t_d_pt

    REAL, DIMENSION(:,:), ALLOCATABLE ::  precipitation_amount, &
                                          precipitation_rate

    SAVE

 END MODULE cloud_parameters




 MODULE constants

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of general constants
!------------------------------------------------------------------------------!

    REAL    ::  pi = 3.141592654

    SAVE

 END MODULE constants




 MODULE control_parameters

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of parameters for program control
!------------------------------------------------------------------------------!

    TYPE plot_precision
       CHARACTER (LEN=6) ::  variable
       INTEGER           ::  precision
    END TYPE plot_precision

    TYPE(plot_precision), DIMENSION(100) ::  plot_3d_precision =               &
                        (/ plot_precision( 'u', 2 ), plot_precision( 'v', 2 ), &
                           plot_precision( 'w', 2 ), plot_precision( 'p', 5 ), &
                           plot_precision( 'pt', 2 ),                          &
                           ( plot_precision( ' ', 1 ), i9 = 1,95 ) /)

    TYPE file_status
       LOGICAL ::  opened, opened_before
    END TYPE file_status

    TYPE(file_status), DIMENSION(200) :: openfile = file_status(.FALSE.,.FALSE.)


    CHARACTER (LEN=1)    ::  cycle_mg = 'w', timestep_reason = ' '
    CHARACTER (LEN=2)    ::  coupling_char = ''
    CHARACTER (LEN=5)    ::  write_binary = 'false'
    CHARACTER (LEN=6)    ::  grid_matching = 'strict'
    CHARACTER (LEN=8)    ::  run_date, run_time
    CHARACTER (LEN=9)    ::  simulated_time_chr
    CHARACTER (LEN=11)   ::  topography_grid_convention = ' '
    CHARACTER (LEN=12)   ::  version = ' ', revision = ' '
    CHARACTER (LEN=16)   ::  conserve_volume_flow_mode = 'default', &
                             loop_optimization = 'default', &
                             momentum_advec = 'pw-scheme', &
                             psolver = 'poisfft', &
                             scalar_advec = 'pw-scheme'
    CHARACTER (LEN=20)   ::  bc_e_b = 'neumann', bc_lr = 'cyclic', &
                             bc_ns = 'cyclic', bc_p_b = 'neumann', &
                             bc_p_t = 'dirichlet', bc_pt_b = 'dirichlet', &
                             bc_pt_t = 'initial_gradient', &
                             bc_q_b = 'dirichlet', bc_q_t = 'neumann', &
                             bc_s_b = 'dirichlet', bc_s_t = 'neumann', &
                             bc_sa_t = 'neumann', &
                             bc_uv_b = 'dirichlet', bc_uv_t = 'dirichlet', &
                             canopy_mode = 'block', &
                             coupling_mode = 'uncoupled', &
                             coupling_mode_remote = 'uncoupled', &
                             dissipation_1d = 'as_in_3d_model', &
                             fft_method = 'system-specific', &
                             mixing_length_1d = 'as_in_3d_model', &
                             random_generator = 'numerical-recipes', &
                             return_addres, return_username, &
                             timestep_scheme = 'runge-kutta-3'
    CHARACTER (LEN=40)   ::  avs_data_file, topography = 'flat'
    CHARACTER (LEN=64)   ::  host = ' '
    CHARACTER (LEN=80)   ::  log_message, run_identifier
    CHARACTER (LEN=100)  ::  initializing_actions = ' ', run_description_header
    CHARACTER (LEN=1000) ::  message_string = ' '

    CHARACTER (LEN=7),  DIMENSION(100) ::  do3d_comp_prec = ' '
    CHARACTER (LEN=10), DIMENSION(10)  ::  data_output_format = ' '
    CHARACTER (LEN=11), DIMENSION(100) ::  data_output = ' ',    &
                                           data_output_user = ' ', doav = ' '
    CHARACTER (LEN=10), DIMENSION(300) ::  data_output_pr = ' '
    CHARACTER (LEN=10), DIMENSION(200) ::  data_output_pr_user = ' '
    CHARACTER (LEN=20), DIMENSION(10)  ::  netcdf_precision = ' '

    CHARACTER (LEN=10), DIMENSION(0:1,100) ::  do2d = ' ', do3d = ' '

    INTEGER ::  abort_mode = 1, average_count_pr = 0, average_count_sp = 0, &
                average_count_3d = 0, current_timestep_number = 0, &
                dist_range = 0, disturbance_level_ind_b, &
                disturbance_level_ind_t, doav_n = 0, dopr_n = 0, &
                dopr_time_count = 0, dopts_time_count = 0, &
                dosp_time_count = 0, dots_time_count = 0, &
                do2d_xy_n = 0, do2d_xz_n = 0, do2d_yz_n = 0, do3d_avs_n = 0, &
                dp_level_ind_b = 0, &
                dvrp_filecount = 0, dz_stretch_level_index, gamma_mg, &
                grid_level, ibc_e_b, ibc_p_b, ibc_p_t, ibc_pt_b, ibc_pt_t, &
                ibc_q_b, ibc_q_t, ibc_sa_t, ibc_uv_b, ibc_uv_t, &
                inflow_disturbance_begin = -1, inflow_disturbance_end = -1, &
                intermediate_timestep_count, intermediate_timestep_count_max, &
                iran = -1234567, last_dt_change = 0, maximum_grid_level, &
                max_pr_user = 0, mgcycles = 0, mg_cycles = -1, &
                mg_switch_to_pe0_level = 0, ngsrb = 2, nsor = 20, &
                nsor_ini = 100, n_sor, normalizing_region = 0, &
                nz_do1d, nz_do3d = -9999, outflow_damping_width = -1, &
                pch_index = 0, prt_time_count = 0, recycling_plane, runnr = 0, &
                skip_do_avs = 0, terminate_coupled = 0, &
                terminate_coupled_remote = 0, timestep_count = 0

    INTEGER ::  dist_nxl(0:1), dist_nxr(0:1), dist_nyn(0:1), dist_nys(0:1), &
                do2d_no(0:1) = 0, do2d_xy_time_count(0:1), &
                do2d_xz_time_count(0:1), do2d_yz_time_count(0:1), &
                do3d_no(0:1) = 0, do3d_time_count(0:1), &
                lad_vertical_gradient_level_ind(10) = -9999, &
                pt_vertical_gradient_level_ind(10) = -9999, &
                q_vertical_gradient_level_ind(10) = -9999, &
                sa_vertical_gradient_level_ind(10) = -9999, &
                section(100,3), section_xy(100) = -9999, &
                section_xz(100) = -9999, section_yz(100) = -9999, &
                ug_vertical_gradient_level_ind(10) = -9999, &
                vg_vertical_gradient_level_ind(10) = -9999

    INTEGER, DIMENSION(:), ALLOCATABLE ::  grid_level_count

    LOGICAL ::  adjust_mixing_length = .FALSE., avs_output = .FALSE., &
                bc_lr_cyc =.TRUE., bc_ns_cyc = .TRUE., &
                call_psolver_at_all_substeps = .TRUE., &
                cloud_droplets = .FALSE., cloud_physics = .FALSE., &
                conserve_volume_flow = .FALSE., constant_diffusion = .FALSE., &
                constant_heatflux = .TRUE., constant_top_heatflux = .TRUE., &
                constant_top_momentumflux = .FALSE., &
                constant_top_salinityflux = .TRUE., &
                constant_waterflux = .TRUE., create_disturbances = .TRUE., &
                cut_spline_overshoot = .TRUE., &
                data_output_2d_on_each_pe = .TRUE., do2d_at_begin = .FALSE., &
                do3d_at_begin = .FALSE., do3d_compress = .FALSE., &
                do_sum = .FALSE., dp_external = .FALSE., dp_smooth = .FALSE., &
                dt_changed = .FALSE., dt_fixed = .FALSE., &
                disturbance_created = .FALSE., &
                first_call_advec_particles = .TRUE., &
                force_print_header = .FALSE., galilei_transformation = .FALSE.,&
                humidity = .FALSE., humidity_remote = .FALSE., &
                inflow_l = .FALSE., inflow_n = .FALSE., &
                inflow_r = .FALSE., inflow_s = .FALSE., iso2d_output = .FALSE.,&
                mg_switch_to_pe0 = .FALSE., &
                netcdf_output = .FALSE., netcdf_64bit = .FALSE., &
                netcdf_64bit_3d = .TRUE., ocean = .FALSE., &
                outflow_l = .FALSE., outflow_n = .FALSE., outflow_r = .FALSE., &
                outflow_s = .FALSE., passive_scalar = .FALSE., &
                plant_canopy = .FALSE., &
                prandtl_layer = .TRUE., precipitation = .FALSE., &
                profil_output = .FALSE., radiation = .FALSE., &
                random_heatflux = .FALSE., run_control_header = .FALSE., &
                run_coupled = .TRUE., sloping_surface = .FALSE., &
                stop_dt = .FALSE., terminate_run = .FALSE., &
                turbulent_inflow = .FALSE., &
                use_prior_plot1d_parameters = .FALSE., use_reference = .FALSE.,&
                use_surface_fluxes = .FALSE., use_top_fluxes = .FALSE., &
                use_ug_for_galilei_tr = .TRUE., use_upstream_for_tke = .FALSE.,&
                wall_adjustment = .TRUE.

    LOGICAL ::  data_output_xy(0:1) = .FALSE., data_output_xz(0:1) = .FALSE., &
                data_output_yz(0:1) = .FALSE.

    REAL ::  advected_distance_x = 0.0, advected_distance_y = 0.0, &
             alpha_surface = 0.0, asselin_filter_factor = 0.1, &
             atmos_ocean_sign = 1.0, &
             averaging_interval = 0.0, averaging_interval_pr = 9999999.9, &
             averaging_interval_sp = 9999999.9, bc_pt_t_val, bc_q_t_val, &
             bottom_salinityflux = 0.0, &
             building_height = 50.0, building_length_x = 50.0, &
             building_length_y = 50.0, building_wall_left = 9999999.9, &
             building_wall_south = 9999999.9, canyon_height = 50.0, &
             canyon_width_x = 9999999.9, canyon_width_y = 9999999.9, &
             canyon_wall_left = 9999999.9, canyon_wall_south = 9999999.9, &
             cthf = 0.0, cfl_factor = -1.0, cos_alpha_surface, &
             coupling_start_time, disturbance_amplitude = 0.25, &
             disturbance_energy_limit = 0.01, &
             disturbance_level_b = -9999999.9, &
             disturbance_level_t = -9999999.9, &
             dp_level_b = 0.0, drag_coefficient = 0.0, &
             dt = -1.0, dt_averaging_input = 0.0, &
             dt_averaging_input_pr = 9999999.9, dt_coupling = 9999999.9, &
             dt_data_output = 9999999.9, &
             dt_data_output_av = 9999999.9, dt_disturb = 9999999.9, &
             dt_dopr = 9999999.9, dt_dopr_listing = 9999999.9, &
             dt_dopts = 9999999.9, dt_dosp = 9999999.9, dt_dots = 9999999.9, &
             dt_do2d_xy = 9999999.9, dt_do2d_xz = 9999999.9, &
             dt_do2d_yz = 9999999.9, dt_do3d = 9999999.9, dt_dvrp = 9999999.9, &
             dt_max = 20.0, dt_prel = 9999999.9, dt_restart = 9999999.9, &
             dt_run_control = 60.0, dt_3d = -1.0, dz = -1.0, &
             dz_max = 9999999.9, dz_stretch_factor = 1.08, &
             dz_stretch_level = 100000.0, e_init = 0.0, e_min = 0.0, &
             end_time = 0.0, &
             f = 0.0, fs = 0.0, g = 9.81, inflow_damping_height = 9999999.9, &
             inflow_damping_width = 9999999.9, kappa = 0.4, km_constant = -1.0,&
             km_damp_max = -1.0, lad_surface = 0.0,  &
             leaf_surface_concentration = 0.0, long_filter_factor = 0.0, &
             maximum_cpu_time_allowed = 0.0, molecular_viscosity = 1.461E-5, &
             old_dt = 1.0E-10, omega = 7.29212E-5, omega_sor = 1.8, &
             overshoot_limit_e = 0.0, overshoot_limit_pt = 0.0, &
             overshoot_limit_u = 0.0, overshoot_limit_v = 0.0, &
             overshoot_limit_w = 0.0, particle_maximum_age = 9999999.9, &
             phi = 55.0, prandtl_number = 1.0, &
             precipitation_amount_interval = 9999999.9, prho_reference, &
             pt_reference = 9999999.9, pt_slope_offset = 0.0, &
             pt_surface = 300.0, pt_surface_initial_change = 0.0, &
             q_surface = 0.0, q_surface_initial_change = 0.0, &
             rayleigh_damping_factor = -1.0, rayleigh_damping_height = -1.0, &
             recycling_width = 9999999.9, residual_limit = 1.0E-4, &
             restart_time = 9999999.9, rho_reference, rho_surface, &
             rif_max = 1.0, rif_min = -5.0, roughness_length = 0.1, &
             sa_surface = 35.0, scalar_exchange_coefficient = 0.0, &
             simulated_time = 0.0, simulated_time_at_begin, sin_alpha_surface, &
             skip_time_data_output = 0.0, skip_time_data_output_av = 9999999.9,&
             skip_time_dopr = 9999999.9, skip_time_dosp = 9999999.9, &
             skip_time_do2d_xy = 9999999.9, skip_time_do2d_xz = 9999999.9, &
             skip_time_do2d_yz = 9999999.9, skip_time_do3d = 9999999.9, &
             surface_heatflux = 9999999.9, surface_pressure = 1013.25, &
             surface_scalarflux = 0.0, surface_waterflux = 0.0, &
             s_surface = 0.0, s_surface_initial_change = 0.0, &
             termination_time_needed = -1.0, time_coupling = 0.0, &
             time_disturb = 0.0, time_dopr = 0.0, time_dopr_av = 0.0, &
             time_dopr_listing = 0.0, time_dopts = 0.0, time_dosp = 0.0, &
             time_dosp_av = 0.0, time_dots = 0.0, time_do2d_xy = 0.0, &
             time_do2d_xz = 0.0, time_do2d_yz = 0.0, time_do3d = 0.0, &
             time_do_av = 0.0, time_do_sla = 0.0, time_dvrp = 0.0, &
             time_prel = 0.0, time_restart = 9999999.9, time_run_control = 0.0,&
             time_since_reference_point, top_heatflux = 9999999.9, &
             top_momentumflux_u = 9999999.9, &
             top_momentumflux_v = 9999999.9, top_salinityflux = 9999999.9, &
             ug_surface = 0.0, u_bulk = 0.0, u_gtrans = 0.0, &
             ups_limit_e = 0.0, ups_limit_pt = 0.0, ups_limit_u = 0.0, &
             ups_limit_v = 0.0, ups_limit_w = 0.0, vg_surface = 0.0, &
             v_bulk = 0.0, v_gtrans = 0.0, wall_adjustment_factor = 1.8, &
             z_max_do1d = -1.0, z_max_do1d_normalized = -1.0, z_max_do2d = -1.0

    REAL ::  do2d_xy_last_time(0:1) = -1.0, do2d_xz_last_time(0:1) = -1.0, &
             do2d_yz_last_time(0:1) = -1.0, dpdxy(1:2) = 0.0, &
             lad_vertical_gradient(10) = 0.0, &
             lad_vertical_gradient_level(10) = -9999999.9, &
             pt_vertical_gradient(10) = 0.0, &
             pt_vertical_gradient_level(10) = -9999999.9, &
             q_vertical_gradient(10) = 0.0, &
             q_vertical_gradient_level(10) = -1.0, &
             s_vertical_gradient(10) = 0.0, &
             s_vertical_gradient_level(10) = -1.0, &
             sa_vertical_gradient(10) = 0.0, &
             sa_vertical_gradient_level(10) = -9999999.9, threshold(20) = 0.0, &
             tsc(10) = (/ 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 /), &
             ug_vertical_gradient(10) = 0.0, &
             ug_vertical_gradient_level(10) = -9999999.9, &
             vg_vertical_gradient(10) = 0.0, &
             vg_vertical_gradient_level(10) = -9999999.9, &
             volume_flow(1:2) = 0.0, volume_flow_area(1:2) = 0.0, &
             volume_flow_initial(1:2) = 0.0, wall_heatflux(0:4) = 0.0, &
             wall_humidityflux(0:4) = 0.0, wall_qflux(0:4) = 0.0, &
             wall_salinityflux(0:4) = 0.0, wall_scalarflux(0:4) = 0.0

    REAL, DIMENSION(:), ALLOCATABLE ::  dp_smooth_factor


    SAVE

 END MODULE control_parameters




 MODULE cpulog

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of variables for cpu-time measurements
!------------------------------------------------------------------------------!

    REAL ::  initial_wallclock_time

    TYPE logpoint
       REAL               ::  isum, ivect, mean, mtime, mtimevec, sum, vector
       INTEGER            ::  counts
       CHARACTER (LEN=20) ::  place
    END TYPE logpoint

    TYPE(logpoint), DIMENSION(100) ::  log_point = logpoint( 0.0, 0.0, 0.0,   &
                                       0.0, 0.0, 0.0, 0.0, 0, ' ' ),          &
                                       log_point_s = logpoint( 0.0, 0.0, 0.0, &
                                       0.0, 0.0, 0.0, 0.0, 0, ' ' )

    SAVE

 END MODULE cpulog




 MODULE dvrp_variables

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of variables used with dvrp-software
!------------------------------------------------------------------------------!

    CHARACTER (LEN=10) ::  dvrp_output = 'rtsp', particle_color = 'none', &
                           particle_dvrpsize = 'none'

    CHARACTER (LEN=20), DIMENSION(10) ::  mode_dvrp = &
                                     (/ ( '                    ', i9 = 1,10 ) /)

    CHARACTER (LEN=80) ::  dvrp_directory = 'default',                    &
                           dvrp_file      = 'default',                    &
                           dvrp_host      = 'origin.rvs.uni-hannover.de', &
                           dvrp_password  = '********',                   &
                           dvrp_username  = ' '

    INTEGER ::  cluster_size = 1, dvrp_colortable_entries = 4,                 &
                dvrp_colortable_entries_prt = 22, islice_dvrp,                 &
                nx_dvrp, nxl_dvrp, nxr_dvrp, ny_dvrp, nyn_dvrp, nys_dvrp,      &
                nz_dvrp, pathlines_fadeintime = 5, pathlines_fadeouttime = 5,  &
                pathlines_linecount = 1000, pathlines_maxhistory = 40,         &
                pathlines_wavecount = 10, pathlines_wavetime = 50,             &
                vc_gradient_normals = 0, vc_mode = 0, vc_size_x = 2,           &
                vc_size_y = 2, vc_size_z = 2

    INTEGER, DIMENSION(10) ::  slicer_position_dvrp

    LOGICAL ::  cyclic_dvrp = .FALSE., dvrp_overlap, dvrp_total_overlap, &
                local_dvrserver_running, lock_steering_update = .FALSE., &
                use_seperate_pe_for_dvrp_output = .FALSE.

    REAL    ::  clip_dvrp_l = 9999999.9, clip_dvrp_n = 9999999.9, &
                clip_dvrp_r = 9999999.9, clip_dvrp_s = 9999999.9, &
                superelevation = 1.0, superelevation_x = 1.0,     &
                superelevation_y = 1.0, vc_alpha = 38.0

    REAL, DIMENSION(2) ::  color_interval = (/ 0.0, 1.0 /), &
                           dvrpsize_interval = (/ 0.0, 1.0 /)

    REAL, DIMENSION(3) ::  groundplate_color = (/ 0.0, 0.6, 0.0 /), &
                           topography_color = (/ 0.8, 0.7, 0.6 /)

#if defined( __decalpha )
    REAL, DIMENSION(2,10)  ::  slicer_range_limits_dvrp = (/ &
                                -1.0, 1.0, -1.0, 1.0, -1.0, 1.0, -1.0, 1.0, &
                                -1.0, 1.0, -1.0, 1.0, -1.0, 1.0, -1.0, 1.0, &
                                -1.0, 1.0, -1.0, 1.0 /)

    REAL, DIMENSION(3,10)  ::  isosurface_color = (/ &
                                0.9, 0.9, 0.9,  0.8, 0.1, 0.1,  0.1, 0.1, 0.8, &
                                0.1, 0.8, 0.1,  0.6, 0.1, 0.1,  0.1, 0.1, 0.6, &
                                0.1, 0.6, 0.1,  0.4, 0.1, 0.1,  0.1, 0.1, 0.4  &
                                0.1, 0.4, 0.1 /)

    REAL(4), DIMENSION(2,100) ::  interval_values_dvrp, interval_h_dvrp =      &
                                  (/ 270.0, 225.0, 225.0, 180.0, 70.0, 25.0,   &
                                     25.0, -25.0, ( 0.0, i9 = 1, 192 ) /),     &
                                  interval_l_dvrp = 0.5, interval_s_dvrp = 1.0,&
                                  interval_a_dvrp = 0.0,                       &
                                  interval_values_dvrp_prt,                    &
                                  interval_h_dvrp_prt,                         &
                                  (/ 270.0, 225.0, 225.0, 180.0, 70.0, 25.0,   &
                                     25.0, -25.0, ( 0.0, i9 = 1, 192 ) /),     &
                                  interval_l_dvrp_prt = 0.5,                   &
                                  interval_s_dvrp_prt = 1.0,                   &
                                  interval_a_dvrp_prt = 0.0
#else
    REAL, DIMENSION(2,10)     ::  slicer_range_limits_dvrp

    REAL, DIMENSION(3,10)     ::  isosurface_color

    REAL(4), DIMENSION(2,100) ::  interval_values_dvrp,                       &
                                  interval_values_dvrp_prt, interval_h_dvrp,  &
                                  interval_h_dvrp_prt, interval_l_dvrp = 0.5, &
                                  interval_l_dvrp_prt = 0.5, interval_s_dvrp = 1.0, &
                                  interval_s_dvrp_prt = 1.0, interval_a_dvrp = 0.0, &
                                  interval_a_dvrp_prt = 0.0

    DATA  slicer_range_limits_dvrp / -1.0, 1.0, -1.0, 1.0, -1.0, 1.0, &
                                     -1.0, 1.0, -1.0, 1.0, -1.0, 1.0, &
                                     -1.0, 1.0, -1.0, 1.0, -1.0, 1.0, &
                                     -1.0, 1.0 /

    DATA  isosurface_color / 0.9, 0.9, 0.9,  0.8, 0.1, 0.1,  0.1, 0.1, 0.8, &
                             0.1, 0.8, 0.1,  0.6, 0.1, 0.1,  0.1, 0.1, 0.6, &
                             0.1, 0.6, 0.1,  0.4, 0.1, 0.1,  0.1, 0.1, 0.4, &
                             0.1, 0.4, 0.1 /

    DATA  interval_h_dvrp / 270.0, 225.0, 225.0, 180.0, 70.0, 25.0, &
                            25.0, -25.0, 192 * 0.0 /

    DATA  interval_h_dvrp_prt / 270.0, 225.0, 225.0, 180.0, 70.0, 25.0, &
                                25.0, -25.0, 192 * 0.0 /
#endif

    REAL(4), DIMENSION(:), ALLOCATABLE ::  xcoor_dvrp, ycoor_dvrp, zcoor_dvrp

    TYPE steering
       CHARACTER (LEN=20) ::  name
       REAL(4)            ::  min, max
       INTEGER            ::  imin, imax
    END TYPE steering

    TYPE(steering), DIMENSION(:), ALLOCATABLE ::  steering_dvrp

    SAVE

 END MODULE dvrp_variables




 MODULE grid_variables

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of grid spacings
!------------------------------------------------------------------------------!

    REAL ::  ddx, ddx2, dx = 1.0, dx2, ddy, ddy2, dy = 1.0, dy2

    REAL, DIMENSION(:), ALLOCATABLE ::  ddx2_mg, ddy2_mg

    REAL, DIMENSION(:,:), ALLOCATABLE ::  fwxm, fwxp, fwym, fwyp, fxm, fxp,   &
                                          fym, fyp, wall_e_x, wall_e_y,       &
                                          wall_u, wall_v, wall_w_x, wall_w_y, &
                                          zu_s_inner, zw_w_inner

    SAVE

 END MODULE grid_variables




 MODULE indices

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of array bounds and number of gridpoints
!------------------------------------------------------------------------------!

    INTEGER ::  ngp_sums, nnx, nx = 0, nxa, nxl, nxlu, nxr, nxra, nx_on_file,  &
                nny, ny = 0, nya, nyn, nyna, nys, nysv, ny_on_file, nnz,       &
                nz = 0, nza, nzb, nzb_diff, nzt, nzta, nzt_diff

    INTEGER, DIMENSION(:), ALLOCATABLE ::                                      &
                ngp_2dh, ngp_3d, ngp_3d_inner,                                 &
                nnx_pe, nny_pe, nxl_mg, nxr_mg, nyn_mg, nys_mg, nzt_mg

    INTEGER, DIMENSION(:,:), ALLOCATABLE ::                                    &
                ngp_2dh_outer, ngp_2dh_s_inner, mg_loc_ind, nzb_diff_s_inner,  &
                nzb_diff_s_outer, nzb_diff_u, nzb_diff_v, nzb_inner, nzb_outer,&
                nzb_s_inner, nzb_s_outer, nzb_u_inner, nzb_u_outer,            &
                nzb_v_inner, nzb_v_outer, nzb_w_inner, nzb_w_outer, nzb_2d

    INTEGER, DIMENSION(:,:,:), POINTER ::  flags

    INTEGER, DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  wall_flags_1,           &
                wall_flags_2, wall_flags_3, wall_flags_4, wall_flags_5,        &
                wall_flags_6, wall_flags_7, wall_flags_8, wall_flags_9,        &
                wall_flags_10

    SAVE

 END MODULE indices




 MODULE interfaces

!------------------------------------------------------------------------------!
! Description:
! ------------
! Interfaces for special subroutines which use optional parameters
!------------------------------------------------------------------------------!

    INTERFACE

       SUBROUTINE cpu_log( log_event, place, modus, barrierwait )

          USE cpulog

          CHARACTER (LEN=*)           ::  modus, place
          CHARACTER (LEN=*), OPTIONAL ::  barrierwait
          TYPE(logpoint)              ::  log_event

       END SUBROUTINE cpu_log

    END INTERFACE



    INTERFACE

       SUBROUTINE global_min_max ( i1, i2, j1, j2, k1, k2, feld, mode, wert, &
                                   wert_ijk, wert1, wert1_ijk )
          CHARACTER (LEN=*), INTENT(IN) ::  mode
          INTEGER, INTENT(IN)           ::  i1, i2, j1, j2, k1, k2
          INTEGER                       ::  wert_ijk(3)
          INTEGER, OPTIONAL             ::  wert1_ijk(3)
          REAL                          ::  wert
          REAL, OPTIONAL                ::  wert1
          REAL, INTENT(IN)              ::  feld(i1:i2,j1:j2,k1:k2)

       END SUBROUTINE global_min_max

    END INTERFACE

    SAVE

 END MODULE interfaces



 MODULE pointer_interfaces

!------------------------------------------------------------------------------!
! Description:
! ------------
! Interfaces for subroutines with pointer arguments called in
! prognostic_equations
!------------------------------------------------------------------------------!

    INTERFACE

       SUBROUTINE advec_s_bc( sk, sk_char )

          CHARACTER (LEN=*), INTENT(IN)   ::  sk_char
          REAL, DIMENSION(:,:,:), POINTER ::  sk

       END SUBROUTINE advec_s_bc

    END INTERFACE


    SAVE

 END MODULE pointer_interfaces




 MODULE model_1d

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of variables for the 1D-model
!------------------------------------------------------------------------------!

    INTEGER ::  current_timestep_number_1d = 0, damp_level_ind_1d, &
                last_dt_change_1d = 0 

    LOGICAL ::  run_control_header_1d = .FALSE., stop_dt_1d = .FALSE. 

    REAL ::     damp_level_1d = -1.0, dt_1d = 60.0, dt_max_1d = 300.0, &
                dt_pr_1d = 9999999.9, dt_run_control_1d = 60.0, &
                end_time_1d = 864000.0, old_dt_1d = 1.0E-10, &
                qs1d, simulated_time_1d = 0.0, time_pr_1d = 0.0, &
                time_run_control_1d = 0.0, ts1d, us1d, usws1d, usws1d_m, &
                vsws1d, vsws1d_m, z01d


    REAL, DIMENSION(:), ALLOCATABLE ::  e1d, e1d_m, e1d_p, kh1d, kh1d_m, km1d, &
                                        km1d_m, l_black, l1d, l1d_m, rif1d,    &
                                        te_e, te_em, te_u, te_um, te_v, te_vm, &
                                        u1d, u1d_m, u1d_p, v1d, v1d_m, v1d_p

    SAVE

 END MODULE model_1d




 MODULE netcdf_control

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of parameters and variables for netcdf control.
!------------------------------------------------------------------------------!

#if defined( __netcdf )
    USE netcdf
#endif

    INTEGER, PARAMETER ::  dopr_norm_num = 7, dopts_num = 26, dots_max = 100, &
                           replace_num = 6

    INTEGER ::  dots_num = 23

    CHARACTER, DIMENSION( replace_num ) :: &
                           replace_char = (/ '''', '"', '*', '/', '(', ')' /), &
                           replace_by   = (/ 'p' , 'p', 's', 'o', '_', '_' /)

    CHARACTER (LEN=6), DIMENSION(dopr_norm_num) ::  dopr_norm_names =   &
         (/ 'wpt0  ', 'ws2   ', 'tsw2  ', 'ws3   ', 'ws2tsw', 'wstsw2', &
            'z_i   ' /)

    CHARACTER (LEN=6), DIMENSION(dopr_norm_num) ::  dopr_norm_longnames = &
         (/ 'wpt0  ', 'w*2   ', 't*w2  ', 'w*3   ', 'w*2t*w', 'w*t*w2',   &
            'z_i   ' /)

    CHARACTER (LEN=7), DIMENSION(dopts_num) :: dopts_label = &
          (/ 'tnpt   ', 'x_     ', 'y_     ', 'z_     ', 'z_abs  ', 'u      ', &
             'v      ', 'w      ', 'u"     ', 'v"     ', 'w"     ', 'npt_up ', &
             'w_up   ', 'w_down ', 'npt_max', 'npt_min', 'x*2    ', 'y*2    ', &
             'z*2    ', 'u*2    ', 'v*2    ', 'w*2    ', 'u"2    ', 'v"2    ', &
             'w"2    ', 'npt*2  ' /)

    CHARACTER (LEN=7), DIMENSION(dopts_num) :: dopts_unit = &
          (/ 'number ', 'm      ', 'm      ', 'm      ', 'm      ', 'm/s    ', &
             'm/s    ', 'm/s    ', 'm/s    ', 'm/s    ', 'm/s    ', 'number ', &
             'm/s    ', 'm/s    ', 'number ', 'number ', 'm2     ', 'm2     ', &
             'm2     ', 'm2/s2  ', 'm2/s2  ', 'm2/s2  ', 'm2/s2  ', 'm2/s2  ', &
             'm2/s2  ', 'number2' /)

    CHARACTER (LEN=7), DIMENSION(dots_max) :: dots_label = &
          (/ 'E      ', 'E*     ', 'dt     ', 'u*     ', 'th*    ', 'umax   ', &
             'vmax   ', 'wmax   ', 'div_new', 'div_old', 'z_i_wpt', 'z_i_pt ', &
             'w*     ', 'w"pt"0 ', 'w"pt"  ', 'wpt    ', 'pt(0)  ', 'pt(zp) ', &
             'w"u"0  ', 'w"v"0  ', 'w"q"0  ', 'mo_L   ', 'q*     ',            &
             ( 'unknown', i9 = 1, 77) /)

    CHARACTER (LEN=7), DIMENSION(dots_max) :: dots_unit = &
          (/ 'm2/s2  ', 'm2/s2  ', 's      ', 'm/s    ', 'K      ', 'm/s    ', &
             'm/s    ', 'm/s    ', 's-1    ', 's-1    ', 'm      ', 'm      ', &
             'm/s    ', 'K m/s  ', 'K m/s  ', 'K m/s  ', 'K      ', 'K      ', &
             'm2/s2  ', 'm2/s2  ', 'kg m/s ', 'm      ', 'kg/kg  ',            &
             ( 'unknown', i9 = 1, 77 ) /)

    CHARACTER (LEN=7), DIMENSION(300) ::  dopr_unit = 'unknown'

    CHARACTER (LEN=7), DIMENSION(0:1,100) ::  do2d_unit, do3d_unit

    CHARACTER (LEN=16), DIMENSION(25) ::  prt_var_names = &
          (/ 'pt_age          ', 'pt_dvrp_size    ', 'pt_origin_x     ', &
             'pt_origin_y     ', 'pt_origin_z     ', 'pt_radius       ', &
             'pt_speed_x      ', 'pt_speed_y      ', 'pt_speed_z      ', &
             'pt_weight_factor', 'pt_x            ', 'pt_y            ', &
             'pt_z            ', 'pt_color        ', 'pt_group        ', &
             'pt_tailpoints   ', 'pt_tail_id      ', 'pt_density_ratio', &
             'pt_exp_arg      ', 'pt_exp_term     ', 'not_used        ', &
             'not_used        ', 'not_used        ', 'not_used        ', &
             'not_used        ' /)

    CHARACTER (LEN=16), DIMENSION(25) ::  prt_var_units = &
          (/ 'seconds         ', 'meters          ', 'meters          ', &
             'meters          ', 'meters          ', 'meters          ', &
             'm/s             ', 'm/s             ', 'm/s             ', &
             'factor          ', 'meters          ', 'meters          ', &
             'meters          ', 'none            ', 'none            ', &
             'none            ', 'none            ', 'ratio           ', &
             'none            ', 'none            ', 'not_used        ', &
             'not_used        ', 'not_used        ', 'not_used        ', &
             'not_used        ' /)

    INTEGER ::  id_dim_prtnum, id_dim_time_pr, id_dim_time_prt, &
                id_dim_time_pts, id_dim_time_sp, id_dim_time_ts, id_dim_x_sp, &
                id_dim_y_sp, id_dim_zu_sp, id_dim_zw_sp, id_set_pr, &
                id_set_prt, id_set_pts, id_set_sp, id_set_ts, id_var_prtnum, &
                id_var_rnop_prt, id_var_time_pr, id_var_time_prt, &
                id_var_time_pts, id_var_time_sp, id_var_time_ts, id_var_x_sp, &
                id_var_y_sp, id_var_zu_sp, id_var_zw_sp, nc_stat

    INTEGER, DIMENSION(0:1) ::  id_dim_time_xy, id_dim_time_xz, &
                id_dim_time_yz, id_dim_time_3d, id_dim_x_xy, id_dim_xu_xy, &
                id_dim_x_xz, id_dim_xu_xz, id_dim_x_yz, id_dim_xu_yz, &
                id_dim_x_3d, id_dim_xu_3d, id_dim_y_xy, id_dim_yv_xy, &
                id_dim_y_xz, id_dim_yv_xz, id_dim_y_yz, id_dim_yv_yz, &
                id_dim_y_3d, id_dim_yv_3d, id_dim_zu_xy, id_dim_zu1_xy, &
                id_dim_zu_xz, id_dim_zu_yz, id_dim_zu_3d, id_dim_zw_xy, &
                id_dim_zw_xz, id_dim_zw_yz, id_dim_zw_3d, id_set_xy, &
                id_set_xz, id_set_yz, id_set_3d, id_var_ind_x_yz, &
                id_var_ind_y_xz, id_var_ind_z_xy, id_var_time_xy, &
                id_var_time_xz, id_var_time_yz, id_var_time_3d, id_var_x_xy, &
                id_var_xu_xy, id_var_x_xz, id_var_xu_xz, id_var_x_yz, &
                id_var_xu_yz, id_var_x_3d, id_var_xu_3d, id_var_y_xy, &
                id_var_yv_xy, id_var_y_xz, id_var_yv_xz, id_var_y_yz, &
                id_var_yv_yz, id_var_y_3d, id_var_yv_3d, id_var_zusi_xy, &
                id_var_zusi_3d, id_var_zu_xy, id_var_zu1_xy, id_var_zu_xz, &
                id_var_zu_yz, id_var_zu_3d, id_var_zwwi_xy, id_var_zwwi_3d, &
                id_var_zw_xy, id_var_zw_xz, id_var_zw_yz, id_var_zw_3d

    INTEGER, DIMENSION(10)  ::  id_var_dospx, id_var_dospy, nc_precision 
    INTEGER, DIMENSION(20)  ::  id_var_prt
    INTEGER, DIMENSION(dopr_norm_num) ::  id_var_norm_dopr

    INTEGER, DIMENSION(dopts_num,0:10) ::  id_var_dopts
    INTEGER, DIMENSION(0:1,100)        ::  id_var_do2d, id_var_do3d
    INTEGER, DIMENSION(100,0:9)        ::  id_dim_z_pr, id_var_dopr, &
                                           id_var_z_pr
    INTEGER, DIMENSION(dots_max,0:9)   ::  id_var_dots

    LOGICAL ::  output_for_t0 = .FALSE.

    SAVE

 END MODULE netcdf_control



 MODULE particle_attributes

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of variables used to compute particle transport
!------------------------------------------------------------------------------!

    USE array_kind

    CHARACTER (LEN=15)  ::  bc_par_lr = 'cyclic',  bc_par_ns = 'cyclic', &
                            bc_par_b  = 'reflect', bc_par_t  = 'absorb'

#if defined( __parallel )
    INTEGER ::  mpi_particle_type
#endif
    INTEGER ::  ibc_par_lr, ibc_par_ns, ibc_par_b, ibc_par_t,                  &
                iran_part = -1234567, maximum_number_of_particles = 1000,      &
                maximum_number_of_tailpoints = 100,                            &
                maximum_number_of_tails = 0,                                   &
                number_of_initial_particles = 0, number_of_particles = 0,      &
                number_of_particle_groups = 1, number_of_tails = 0,            &
                number_of_initial_tails = 0, offset_ocean_nzt = 0,             &
                offset_ocean_nzt_m1 = 0, particles_per_point = 1,              &
                particle_file_count = 0, skip_particles_for_tail = 100,        &
                total_number_of_particles, total_number_of_tails = 0

    INTEGER, PARAMETER ::  max_number_of_particle_groups = 10

    INTEGER, DIMENSION(:), ALLOCATABLE     ::  new_tail_id
    INTEGER, DIMENSION(:,:,:), ALLOCATABLE ::  prt_count, prt_start_index

    LOGICAL ::  particle_advection = .FALSE., random_start_position = .FALSE., &
                read_particles_from_restartfile = .TRUE.,                      &
                uniform_particles = .TRUE., use_particle_tails = .FALSE.,      &
                use_sgs_for_particles = .FALSE.,                               &
                write_particle_statistics = .FALSE.

    LOGICAL, DIMENSION(max_number_of_particle_groups) ::                       &
                vertical_particle_advection = .TRUE.

    LOGICAL, DIMENSION(:), ALLOCATABLE ::  particle_mask, tail_mask

    REAL    ::  c_0 = 3.0, dt_min_part = 0.0002, dt_sort_particles = 0.0,      &
                dt_write_particle_data = 9999999.9, dvrp_psize = 9999999.9,    &
                end_time_prel = 9999999.9, initial_weighting_factor = 1.0,     &
                maximum_tailpoint_age = 100000.0,                              &
                minimum_tailpoint_distance = 0.0,                              &
                particle_advection_start = 0.0, sgs_wfu_part = 0.3333333,      &
                sgs_wfv_part = 0.3333333, sgs_wfw_part = 0.3333333,            &
                time_sort_particles = 0.0, time_write_particle_data = 0.0

    REAL, DIMENSION(max_number_of_particle_groups) ::  &
                density_ratio = 9999999.9, pdx = 9999999.9, pdy = 9999999.9, &
                pdz = 9999999.9, psb = 9999999.9, psl = 9999999.9,           &
                psn = 9999999.9, psr = 9999999.9, pss = 9999999.9,           &
                pst = 9999999.9, radius = 9999999.9

    REAL, DIMENSION(:,:,:), ALLOCATABLE ::  particle_tail_coordinates


    TYPE particle_type
       SEQUENCE
       REAL    ::  age, age_m, dt_sum, dvrp_psize, e_m, origin_x, origin_y, &
                   origin_z, radius, speed_x, speed_x_sgs, speed_y,         &
                   speed_y_sgs, speed_z, speed_z_sgs, weight_factor, x, y, z
       INTEGER ::  color, group, tailpoints, tail_id
    END TYPE particle_type

    TYPE(particle_type), DIMENSION(:), ALLOCATABLE ::  initial_particles, &
                                                       particles

    TYPE particle_groups_type
       SEQUENCE
       REAL    ::  density_ratio, radius, exp_arg, exp_term
    END TYPE particle_groups_type

    TYPE(particle_groups_type), DIMENSION(max_number_of_particle_groups) ::&
                   particle_groups

    SAVE

 END MODULE particle_attributes





 MODULE pegrid

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of variables which define processor topology and the exchange of
! ghost point layers. This modules must be placed in all routines which contain
! MPI-calls.
!------------------------------------------------------------------------------!

#if defined( __parallel )
#if defined( __lc )
    USE MPI
#else
    INCLUDE "mpif.h"
#endif
#endif
    CHARACTER(LEN=5)       ::  myid_char = ''
    INTEGER                ::  id_inflow = 0, id_recycling = 0, myid = 0,      &
                               target_id, npex = -1, npey = -1, numprocs = 1,  &
                               numprocs_previous_run = -1,                     &
                               tasks_per_node = -9999, threads_per_task = 1

    INTEGER, DIMENSION(:,:), ALLOCATABLE ::  hor_index_bounds, &
                                             hor_index_bounds_previous_run

#if defined( __parallel )
#if defined( __mpi2 )
    CHARACTER (LEN=MPI_MAX_PORT_NAME) ::  port_name
#endif

    INTEGER ::  comm1dx, comm1dy, comm2d, comm_inter, comm_palm, ierr, myidx,  &
                myidy, ndim = 2, ngp_xy, ngp_y, pleft, pnorth, pright, psouth, &
                sendrecvcount_xy, sendrecvcount_yz, sendrecvcount_zx,          &
                sendrecvcount_zyd, sendrecvcount_yxd,                          &
                type_x, type_x_int, type_xy

    INTEGER ::  ibuf(12), pcoord(2), pdims(2), status(MPI_STATUS_SIZE)

    INTEGER, DIMENSION(:), ALLOCATABLE ::  ngp_yz, type_xz

    LOGICAL ::  reorder = .TRUE.
    LOGICAL, DIMENSION(2) ::  cyclic = (/ .TRUE. , .TRUE. /), &
                              remain_dims
#endif

    SAVE

 END MODULE pegrid




 MODULE profil_parameter

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of variables which control PROFIL-output
!------------------------------------------------------------------------------!

    INTEGER, PARAMETER ::  crmax = 100

    CHARACTER (LEN=10), DIMENSION(100) ::  dopr_label = ' '

    CHARACTER (LEN=10), DIMENSION(crmax) ::  cross_normalized_x = ' ', &
                                             cross_normalized_y = ' '

    CHARACTER (LEN=20), DIMENSION(20) ::  cross_ts_profiles = &
                           (/ ' E E*               ', ' dt                 ', &
                              ' u* w*              ', ' th*                ', &
                              ' umax vmax wmax     ', ' div_old div_new    ', &
                              ' z_i_wpt z_i_pt     ', ' w"pt"0 w"pt" wpt   ', &
                              ' pt(0) pt(zp)       ', ' splux spluy spluz  ', &
                              ' L                  ',                         &
                            ( '                    ', i9 = 1, 9 ) /)

    CHARACTER (LEN=40), DIMENSION(crmax) ::  cross_xtext = &
                           (/ 'windspeed in ms>->1         ', &
                              'pot. temperature in K       ', &
                              'heat flux in K ms>->1       ', &
                              'momentum flux in m>2s>2     ', &
                              'eddy diffusivity in m>2s>->1', &
                              'mixing length in m          ', &
                            ( '                            ', i9 = 1, 94 ) /)

    CHARACTER (LEN=100), DIMENSION(crmax) ::  cross_profiles = &
                           (/ ' u v                           ', &
                              ' pt                            ', &
                              ' w"pt" w*pt* w*pt*BC wpt wptBC ', &
                              ' w"u" w*u* wu w"v" w*v* wv     ', &
                              ' km kh                         ', &
                              ' l                             ', &
                         ( '                               ', i9 = 1, 94 ) /)

    INTEGER ::  profile_columns = 3, profile_rows = 2, profile_number = 0

    INTEGER ::  cross_linecolors(100,crmax) = 1, &
                cross_linestyles(100,crmax) = 0, &
                cross_profile_numbers(100,crmax) = 0, &
                cross_pnc_local(crmax), cross_profile_number_count(crmax) = 0, &
                cross_ts_numbers(crmax,crmax) = 0, &
                cross_ts_number_count(crmax) = 0, dopr_crossindex(100) = 0, &
                dopr_index(300) = 0, dopr_initial_index(300) = 0, &
                dots_crossindex(100) = 0, dots_index(100) = 0, &
                linecolors(10) = (/ 2, 3, 4,  5, 7, 8, 12, 15, 16, 23 /), &
                linestyles(11) = (/ 0, 7, 3, 10, 4, 1,  9,  2,  5,  8, 6 /)
                

    REAL ::  cross_normx_factor(100,crmax) = 1.0, &
             cross_normy_factor(100,crmax) = 1.0, &
             cross_ts_uymax(20) = &
                             (/ 999.999, 999.999, 999.999, 999.999, 999.999,   &
                                999.999, 999.999, 999.999, 999.999, 999.999,   &
                                999.999, 999.999, 999.999, 999.999, 999.999,   &
                                999.999, 999.999, 999.999, 999.999, 999.999 /),&
             cross_ts_uymax_computed(20) = 999.999, &
             cross_ts_uymin(20) = &
                             (/ 999.999, 999.999, 999.999,  -5.000, 999.999,   &
                                999.999,   0.000, 999.999, 999.999, 999.999,   &
                                999.999, 999.999, 999.999, 999.999, 999.999,   &
                                999.999, 999.999, 999.999, 999.999, 999.999 /),&
             cross_ts_uymin_computed(20) = 999.999, &
             cross_uxmax(crmax) = 0.0, cross_uxmax_computed(crmax) = -1.0, &
             cross_uxmax_normalized(crmax) = 0.0, &
             cross_uxmax_normalized_computed(crmax) = -1.0, &
             cross_uxmin(crmax) = 0.0, cross_uxmin_computed(crmax) =  1.0, &
             cross_uxmin_normalized(crmax) = 0.0, &
             cross_uxmin_normalized_computed(crmax) = 1.0, &
             cross_uymax(crmax), cross_uymin(crmax)

    SAVE

 END MODULE profil_parameter




 MODULE spectrum

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of quantities used for computing spectra
!------------------------------------------------------------------------------!

    CHARACTER (LEN=6),  DIMENSION(1:5) ::  header_char = (/ 'PS(u) ', 'PS(v) ',&
                                           'PS(w) ', 'PS(pt)', 'PS(q) ' /)
    CHARACTER (LEN=2),  DIMENSION(10)  ::  spectra_direction = 'x'
    CHARACTER (LEN=10), DIMENSION(10)  ::  data_output_sp  = ' '
    CHARACTER (LEN=25), DIMENSION(1:5) ::  utext_char =                    &
                                           (/ '-power spectrum of u     ', &
                                              '-power spectrum of v     ', &
                                              '-power spectrum of w     ', &
                                              '-power spectrum of ^1185 ', &
                                              '-power spectrum of q     ' /)
    CHARACTER (LEN=39), DIMENSION(1:5) ::  ytext_char =                        &
                                 (/ 'k ^2236 ^2566^2569<u(k) in m>2s>->2    ', &
                                    'k ^2236 ^2566^2569<v(k) in m>2s>->2    ', &
                                    'k ^2236 ^2566^2569<w(k) in m>2s>->2    ', &
                                    'k ^2236 ^2566^2569<^1185(k) in m>2s>->2', &
                                    'k ^2236 ^2566^2569<q(k) in m>2s>->2    ' /)

    INTEGER ::  klist_x = 0, klist_y = 0, n_sp_x = 0, n_sp_y = 0

    INTEGER ::  comp_spectra_level(100) = 999999,                   &
                lstyles(100) = (/ 0, 7, 3, 10, 1, 4, 9, 2, 6, 8,    &
                                  0, 7, 3, 10, 1, 4, 9, 2, 6, 8,    &
                                  0, 7, 3, 10, 1, 4, 9, 2, 6, 8,    &
                                  0, 7, 3, 10, 1, 4, 9, 2, 6, 8,    &
                                  0, 7, 3, 10, 1, 4, 9, 2, 6, 8,    &
                                  0, 7, 3, 10, 1, 4, 9, 2, 6, 8,    &
                                  0, 7, 3, 10, 1, 4, 9, 2, 6, 8,    &
                                  0, 7, 3, 10, 1, 4, 9, 2, 6, 8,    &
                                  0, 7, 3, 10, 1, 4, 9, 2, 6, 8,    &
                                  0, 7, 3, 10, 1, 4, 9, 2, 6, 8 /), &
                plot_spectra_level(100) = 999999

    REAL    ::  time_to_start_sp = 0.0

    SAVE

 END MODULE spectrum




 MODULE statistics

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of statistical quantities, e.g. global sums
!------------------------------------------------------------------------------!

    CHARACTER (LEN=40) ::  region(0:9)
    INTEGER ::  pr_palm = 80, statistic_regions = 0, var_ts = 100
    INTEGER ::  u_max_ijk(3), v_max_ijk(3), w_max_ijk(3)
    LOGICAL ::  flow_statistics_called = .FALSE.
    REAL ::     u_max, v_max, w_max
    REAL, DIMENSION(:), ALLOCATABLE       ::  sums_divnew_l, sums_divold_l
    REAL, DIMENSION(:,:), ALLOCATABLE     ::  sums, sums_wsts_bc_l, ts_value
    REAL, DIMENSION(:,:,:), ALLOCATABLE   ::  hom_sum, rmask, spectrum_x, &
                                              spectrum_y, sums_l, sums_l_l, &
                                              sums_up_fraction_l
    REAL, DIMENSION(:,:,:,:), ALLOCATABLE ::  hom

    SAVE

 END MODULE statistics




 MODULE transpose_indices

!------------------------------------------------------------------------------!
! Description:
! ------------
! Definition of indices for transposed arrays
!------------------------------------------------------------------------------!

    INTEGER ::  nxl_y, nxl_yd, nxl_z, nxr_y, nxr_ya, nxr_yd, nxr_yda, nxr_z, &
                nxr_za, nyn_x, nyn_xa, nyn_z, nyn_za, nys_x, nys_z, nzb_x,   &
                nzb_y, nzb_yd, nzt_x, nzt_xa, nzt_y, nzt_ya, nzt_yd, nzt_yda
                

    SAVE

 END MODULE transpose_indices
